import objax
from jax import vmap, grad, jacrev
import jax.numpy as np
from jax.scipy.linalg import cholesky, cho_factor
from .utils import inv, solve, gaussian_first_derivative_wrt_mean, gaussian_second_derivative_wrt_mean, transpose
from numpy.polynomial.hermite import hermgauss
import numpy as onp
import itertools


class Cubature(objax.Module):

    def __init__(self, dim=None):
        if dim is None:  # dimension of cubature not known upfront
            self.store = False
        else:  # dimension known, store sigma points and weights
            self.store = True
            self.x, self.w = self.get_cubature_points_and_weights(dim)

    def __call__(self, dim):
        if self.store:
            return self.x, self.w
        else:
            return self.get_cubature_points_and_weights(dim)

    def get_cubature_points_and_weights(self, dim):
        raise NotImplementedError


class GaussHermite(Cubature):

    def __init__(self, dim=None, num_cub_points=20):
        self.num_cub_points = num_cub_points
        super().__init__(dim)

    def get_cubature_points_and_weights(self, dim):
        return gauss_hermite(dim, self.num_cub_points)


class UnscentedThirdOrder(Cubature):

    def get_cubature_points_and_weights(self, dim):
        return symmetric_cubature_third_order(dim)


class UnscentedFifthOrder(Cubature):

    def get_cubature_points_and_weights(self, dim):
        return symmetric_cubature_fifth_order(dim)


class Unscented(UnscentedFifthOrder):
    pass


def mvhermgauss(H: int, D: int):
    """
    This function is adapted from GPflow: https://github.com/GPflow/GPflow

    Return the evaluation locations 'xn', and weights 'wn' for a multivariate
    Gauss-Hermite quadrature.

    The outputs can be used to approximate the following type of integral:
    int exp(-x)*f(x) dx ~ sum_i w[i,:]*f(x[i,:])

    :param H: Number of Gauss-Hermite evaluation points.
    :param D: Number of input dimensions. Needs to be known at call-time.
    :return: eval_locations 'x' (H**DxD), weights 'w' (H**D)
    """
    gh_x, gh_w = hermgauss(H)
    x = np.array(list(itertools.product(*(gh_x,) * D)))  # H**DxD
    w = np.prod(np.array(list(itertools.product(*(gh_w,) * D))), 1)  # H**D
    return x, w


def gauss_hermite(dim=1, num_quad_pts=20):
    """
    Return weights and sigma-points for Gauss-Hermite cubature
    """
    # sigma_pts, weights = hermgauss(num_quad_pts)  # Gauss-Hermite sigma points and weights
    sigma_pts, weights = mvhermgauss(num_quad_pts, dim)
    sigma_pts = np.sqrt(2) * sigma_pts.T
    weights = weights.T * np.pi ** (-0.5 * dim)  # scale weights by 1/√π
    return sigma_pts, weights


def symmetric_cubature_third_order(dim=1, kappa=None):
    """
    Return weights and sigma-points for the symmetric cubature rule of order 3.
    Uses 2dim+1 sigma-points
    """
    if kappa is None:
        # kappa = 1 - dim
        kappa = 0  # CKF
    w0 = kappa / (dim + kappa)
    wm = 1 / (2 * (dim + kappa))
    u = onp.sqrt(dim + kappa)
    if (dim == 1) and (kappa == 0):
        weights = onp.array([w0, wm, wm])
        sigma_pts = onp.array([0., u, -u])
        # sigma_pts = onp.array([-u, 0., u])
        # weights = onp.array([wm, w0, wm])
    elif (dim == 2) and (kappa == 0):
        weights = onp.array([w0, wm, wm, wm, wm])
        sigma_pts = onp.block([[0., u,  0., -u, 0.],
                               [0., 0., u, 0., -u]])
    elif (dim == 3) and (kappa == 0):
        weights = onp.array([w0, wm, wm, wm, wm, wm, wm])
        sigma_pts = onp.block([[0., u,  0., 0., -u,   0.,  0.],
                               [0., 0., u,  0.,  0., -u,   0.],
                               [0., 0., 0., u,   0.,  0., -u]])
    else:
        weights = onp.concatenate([onp.array([[kappa / (dim + kappa)]]), wm * onp.ones([1, 2*dim])], axis=1)
        sigma_pts = onp.sqrt(dim + kappa) * onp.block([onp.zeros([dim, 1]), onp.eye(dim), -onp.eye(dim)])
    return sigma_pts, weights


def symmetric_cubature_fifth_order(dim=1):
    """
    Return weights and sigma-points for the symmetric cubature rule of order 5
    Uses 2(dim**2)+1 sigma-points
    """
    # The weights and sigma-points from McNamee & Stenger
    I0 = 1
    I2 = 1
    I4 = 3
    I22 = 1
    u = onp.sqrt(I4 / I2)
    A0 = I0 - dim * (I2 / I4) ** 2 * (I4 - 0.5 * (dim - 1) * I22)
    A1 = 0.5 * (I2 / I4) ** 2 * (I4 - (dim - 1) * I22)
    A2 = 0.25 * (I2 / I4) ** 2 * I22
    # we implement specific cases manually to save compute
    if dim == 1:
        weights = onp.array([A0, A1, A1])
        sigma_pts = onp.array([0., u, -u])
    elif dim == 2:
        weights = onp.array([A0, A1, A1, A1, A1, A2, A2, A2, A2])
        sigma_pts = onp.block([[0., u, -u, 0., 0., u, -u, u, -u],
                               [0., 0., 0., u, -u, u, -u, -u, u]])
    elif dim == 3:
        weights = onp.array([A0, A1, A1, A1, A1, A1, A1, A2, A2, A2, A2, A2, A2, A2, A2, A2, A2, A2, A2])
        sigma_pts = onp.block([[0., u, -u, 0., 0., 0., 0., u, -u, u, -u, u, -u, u, -u, 0., 0., 0., 0.],
                               [0., 0., 0., u, -u, 0., 0., u, -u, -u, u, 0., 0., 0., 0., u, -u, u, -u],
                               [0., 0., 0., 0., 0., u, -u, 0., 0., 0., 0., u, -u, -u, u, u, -u, -u, u]])
    else:
        # general case
        U0 = sym_set(dim, [])
        U1 = sym_set(dim, [u])
        U2 = sym_set(dim, [u, u])

        sigma_pts = onp.concatenate([U0, U1, U2], axis=1)
        weights = onp.concatenate([A0 * onp.ones(U0.shape[1]),
                                   A1 * onp.ones(U1.shape[1]),
                                   A2 * onp.ones(U2.shape[1])])

    return sigma_pts, weights


def sym_set(n, gen=None):

    if (gen is None) or (len(gen) == 0):
        U = onp.zeros([n, 1])

    else:
        lengen = len(gen)
        if lengen == 1:
            U = onp.zeros([n, 2 * n])
        elif lengen == 2:
            U = onp.zeros([n, 2 * n * (n - 1)])
        else:
            raise NotImplementedError

        ind = 0
        for i in range(n):
            u = onp.zeros(n)
            u[i] = gen[0]
            if lengen > 1:
                if abs(gen[0] - gen[1]) < 1e-10:
                    V = sym_set(n-i-1, gen[1:])
                    for j in range(V.shape[1]):
                        u[i+1:] = V[:, j]
                        U[:, 2*ind] = u
                        U[:, 2*ind + 1] = -u
                        ind += 1
                else:
                    raise NotImplementedError
                    # V = sym_set(n-1, gen[1:])
                    # for j in range(V.shape[1]):
                    #     u[:i-1, i+1:] = V[:, j]
                    #     U = onp.concatenate([U, u, -u])
                    #     ind += 1
            else:
                U[:, 2*i] = u
                U[:, 2*i+1] = -u
    return U


def variational_expectation_cubature(likelihood, y, post_mean, post_cov, cubature=None):
    """
    Computes the "variational expectation" via cubature, i.e. the
    expected log-likelihood, and its derivatives w.r.t. the posterior mean
        E[log p(yₙ|fₙ)] = ∫ log p(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    with EP power a.
    :param likelihood: the likelihood model
    :param y: observed data (yₙ) [scalar]
    :param post_mean: posterior mean (mₙ) [scalar]
    :param post_cov: posterior variance (vₙ) [scalar]
    :param cubature: the function to compute sigma points and weights to use during cubature
    :return:
        exp_log_lik: the expected log likelihood, E[log p(yₙ|fₙ)]  [scalar]
        dE_dm: derivative of E[log p(yₙ|fₙ)] w.r.t. mₙ  [scalar]
        d2E_dm2: second derivative of E[log p(yₙ|fₙ)] w.r.t. mₙ  [scalar]
    """
    if cubature is None:
        x, w = gauss_hermite(post_mean.shape[0])  # Gauss-Hermite sigma points and weights
    else:
        x, w = cubature(post_mean.shape[0])
    post_cov = (post_cov + post_cov.T) / 2
    # fsigᵢ=xᵢ√(vₙ) + mₙ: scale locations according to cavity dist.
    sigma_points = cholesky(post_cov, lower=True) @ np.atleast_2d(x) + post_mean
    # pre-compute wᵢ log p(yₙ|xᵢ√(2vₙ) + mₙ)
    weighted_log_likelihood_eval = w * likelihood.evaluate_log_likelihood(y, sigma_points)
    # Compute expected log likelihood via cubature:
    # E[log p(yₙ|fₙ)] = ∫ log p(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    #                 ≈ ∑ᵢ wᵢ log p(yₙ|fsigᵢ)
    exp_log_lik = np.sum(
        weighted_log_likelihood_eval
    )
    # Compute first derivative via cubature:
    # dE[log p(yₙ|fₙ)]/dmₙ = ∫ (fₙ-mₙ) vₙ⁻¹ log p(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    #                      ≈ ∑ᵢ wᵢ (fₙ-mₙ) vₙ⁻¹ log p(yₙ|fsigᵢ)
    invv = np.diag(post_cov)[:, None] ** -1
    dE_dm = np.sum(
        invv * (sigma_points - post_mean)
        * weighted_log_likelihood_eval, axis=-1
    )[:, None]
    # Compute second derivative via cubature (deriv. w.r.t. var = 0.5 * 2nd deriv. w.r.t. mean):
    # dE[log p(yₙ|fₙ)]/dvₙ = ∫ [(fₙ-mₙ)² vₙ⁻² - vₙ⁻¹]/2 log p(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    #                        ≈ ∑ᵢ wᵢ [(fₙ-mₙ)² vₙ⁻² - vₙ⁻¹]/2 log p(yₙ|fsigᵢ)
    dE_dv = np.sum(
        (0.5 * (invv ** 2 * (sigma_points - post_mean) ** 2) - 0.5 * invv)
        * weighted_log_likelihood_eval, axis=-1
    )
    dE_dv = np.diag(dE_dv)
    d2E_dm2 = 2 * dE_dv
    return exp_log_lik, dE_dm, d2E_dm2


def log_density_cubature(likelihood, y, mean, cov, cubature=None):
    """
    logZₙ = log ∫ p(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    :param likelihood: the likelihood model
    :param y: observed data (yₙ) [scalar]
    :param mean: cavity mean (mₙ) [scalar]
    :param cov: cavity covariance (cₙ) [scalar]
    :param cubature: the function to compute sigma points and weights to use during cubature
    :return:
        lZ: the log density, logZₙ  [scalar]
    """
    if cubature is None:
        x, w = gauss_hermite(mean.shape[0])  # Gauss-Hermite sigma points and weights
    else:
        x, w = cubature(mean.shape[0])
    cov = (cov + cov.T) / 2
    cav_cho, low = cho_factor(cov, lower=True)
    # fsigᵢ=xᵢ√cₙ + mₙ: scale locations according to cavity dist.
    sigma_points = cav_cho @ np.atleast_2d(x) + mean
    # pre-compute wᵢ p(yₙ|xᵢ√(2vₙ) + mₙ)
    weighted_likelihood_eval = w * likelihood.evaluate_likelihood(y, sigma_points)
    # Compute partition function via cubature:
    # Zₙ = ∫ p(yₙ|fₙ) 𝓝(fₙ|mₙ,vₙ) dfₙ ≈ ∑ᵢ wᵢ p(yₙ|fsigᵢ)
    Z = np.sum(
        weighted_likelihood_eval
    )
    lZ = np.log(np.maximum(Z, 1e-8))
    return lZ


def log_density_power_cubature(likelihood, y, mean, cov, power=1., cubature=None):
    """
    logZₙ = log ∫ p^a(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    :param likelihood: the likelihood model
    :param y: observed data (yₙ) [scalar]
    :param mean: cavity mean (mₙ) [scalar]
    :param cov: cavity covariance (cₙ) [scalar]
    :param power: EP power [scalar]
    :param cubature: the function to compute sigma points and weights to use during cubature
    :return:
        lZ: the log density, logZₙ  [scalar]
    """
    if cubature is None:
        x, w = gauss_hermite(mean.shape[0])  # Gauss-Hermite sigma points and weights
    else:
        x, w = cubature(mean.shape[0])
    cov = (cov + cov.T) / 2
    cav_cho, low = cho_factor(cov, lower=True)
    # fsigᵢ=xᵢ√cₙ + mₙ: scale locations according to cavity dist.
    sigma_points = cav_cho @ np.atleast_2d(x) + mean
    # pre-compute wᵢ p(yₙ|xᵢ√(2vₙ) + mₙ)
    weighted_likelihood_eval = w * np.exp(power * likelihood.evaluate_log_likelihood(y, sigma_points))
    # Compute partition function via cubature:
    # Zₙ = ∫ p^a(yₙ|fₙ) 𝓝(fₙ|mₙ,vₙ) dfₙ ≈ ∑ᵢ wᵢ p^a(yₙ|fsigᵢ)
    Z = np.sum(
        weighted_likelihood_eval
    )
    lZ = np.log(Z)
    return lZ


def moment_match_cubature(likelihood, y, cav_mean, cav_cov, power=1.0, cubature=None):
    """
    TODO: N.B. THIS VERSION ALLOWS MULTI-DIMENSIONAL MOMENT MATCHING, BUT CAN BE UNSTABLE
    Perform moment matching via cubature.
    Moment matching involves computing the log partition function, logZₙ, and its derivatives w.r.t. the cavity mean
        logZₙ = log ∫ pᵃ(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    with EP power a.
    :param likelihood: the likelihood model
    :param y: observed data (yₙ) [scalar]
    :param cav_mean: cavity mean (mₙ) [scalar]
    :param cav_cov: cavity covariance (cₙ) [scalar]
    :param power: EP power / fraction (a) [scalar]
    :param cubature: the function to compute sigma points and weights to use during cubature
    :return:
        lZ: the log partition function, logZₙ  [scalar]
        dlZ: first derivative of logZₙ w.r.t. mₙ (if derivatives=True)  [scalar]
        d2lZ: second derivative of logZₙ w.r.t. mₙ (if derivatives=True)  [scalar]
    """
    if cubature is None:
        x, w = gauss_hermite(cav_mean.shape[0])  # Gauss-Hermite sigma points and weights
    else:
        x, w = cubature(cav_mean.shape[0])
    cav_cov = (cav_cov + cav_cov.T) / 2
    cav_cho, low = cho_factor(cav_cov, lower=True)
    # fsigᵢ=xᵢ√cₙ + mₙ: scale locations according to cavity dist.
    sigma_points = cav_cho @ np.atleast_2d(x) + cav_mean
    # pre-compute wᵢ pᵃ(yₙ|xᵢ√(2vₙ) + mₙ)
    weighted_likelihood_eval = w * np.exp(power * likelihood.evaluate_log_likelihood(y, sigma_points))
    weighted_likelihood_eval = np.atleast_2d(weighted_likelihood_eval)

    # Compute partition function via cubature:
    # Zₙ = ∫ pᵃ(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    #    ≈ ∑ᵢ wᵢ pᵃ(yₙ|fsigᵢ)
    Z = np.sum(
        weighted_likelihood_eval
    )
    lZ = np.log(np.maximum(Z, 1e-8))
    Zinv = 1.0 / np.maximum(Z, 1e-8)

    # Compute derivative of partition function via cubature:
    # dZₙ/dmₙ = ∫ (fₙ-mₙ) vₙ⁻¹ pᵃ(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    #         ≈ ∑ᵢ wᵢ (fₙ-mₙ) vₙ⁻¹ pᵃ(yₙ|fsigᵢ)
    d1 = vmap(
        gaussian_first_derivative_wrt_mean, (1, None, None, 1)
    )(sigma_points[..., None], cav_mean, cav_cov, weighted_likelihood_eval)
    dZ = np.sum(d1, axis=0)
    # dlogZₙ/dmₙ = (dZₙ/dmₙ) / Zₙ
    dlZ = Zinv * dZ

    # Compute second derivative of partition function via cubature:
    # d²Zₙ/dmₙ² = ∫ [(fₙ-mₙ)² vₙ⁻² - vₙ⁻¹] pᵃ(yₙ|fₙ) N(fₙ|mₙ,vₙ) dfₙ
    #           ≈ ∑ᵢ wᵢ [(fₙ-mₙ)² vₙ⁻² - vₙ⁻¹] pᵃ(yₙ|fsigᵢ)
    d2 = vmap(
        gaussian_second_derivative_wrt_mean, (1, None, None, 1)
    )(sigma_points[..., None], cav_mean, cav_cov, weighted_likelihood_eval)
    d2Z = np.sum(d2, axis=0)

    # d²logZₙ/dmₙ² = d[(dZₙ/dmₙ) / Zₙ]/dmₙ
    #              = (d²Zₙ/dmₙ² * Zₙ - (dZₙ/dmₙ)²) / Zₙ²
    #              = d²Zₙ/dmₙ² / Zₙ - (dlogZₙ/dmₙ)²
    d2lZ = -dlZ @ dlZ.T + Zinv * d2Z
    return lZ, dlZ, d2lZ


def statistical_linear_regression_cubature(likelihood, mean, cov, cubature=None):
    """
    Perform statistical linear regression (SLR) using cubature.
    We aim to find a likelihood approximation p(yₙ|fₙ) ≈ N(yₙ|Afₙ+b,Ω).
    """
    mu, omega = expected_conditional_mean_cubature(likelihood, mean, cov, cubature)
    dmu_dm = expected_conditional_mean_dm(likelihood, mean, cov, cubature)
    d2mu_dm2 = expected_conditional_mean_dm2(likelihood, mean, cov, cubature)
    return mu.reshape(-1, 1), omega, np.atleast_2d(dmu_dm), d2mu_dm2
    # return mu.reshape(-1, 1), omega, dmu_dm[None], np.swapaxes(d2mu_dm2, axis1=0, axis2=2)


def expected_conditional_mean_cubature(likelihood, mean, cov, cubature=None):
    """
    Compute Eq[E[y|f]] = ∫ Ey[p(y|f)] N(f|mean,cov) dfₙ
    """
    if cubature is None:
        x, w = gauss_hermite(mean.shape[0])  # Gauss-Hermite sigma points and weights
    else:
        x, w = cubature(mean.shape[0])
    cov = (cov + cov.T) / 2
    # fsigᵢ=xᵢ√(vₙ) + mₙ: scale locations according to cavity dist.
    sigma_points = cholesky(cov, lower=True) @ np.atleast_2d(x) + mean
    lik_expectation, lik_covariance = likelihood.conditional_moments(sigma_points)
    # Compute muₙ via cubature:
    # muₙ = ∫ E[yₙ|fₙ] N(fₙ|mₙ,vₙ) dfₙ
    #    ≈ ∑ᵢ wᵢ E[yₙ|fsigᵢ]
    mu = np.sum(
        w * lik_expectation, axis=-1
    )[:, None]
    diff = (lik_expectation - mu).T[..., None]
    quadratic_part = diff @ transpose(diff)
    S = np.sum(
        # w * ((lik_expectation - mu) ** 2 + lik_covariance), axis=-1
        w * (quadratic_part.T + lik_covariance), axis=-1
    )
    # Compute cross covariance C via cubature:
    # C = ∫ (fₙ-mₙ) (E[yₙ|fₙ]-muₙ)' N(fₙ|mₙ,vₙ) dfₙ
    #   ≈ ∑ᵢ wᵢ (fsigᵢ -mₙ) (E[yₙ|fsigᵢ]-muₙ)'
    diff2 = (sigma_points - mean).T[..., None]
    quadratic_part2 = diff2 @ transpose(diff)
    C = np.sum(
        # w * (sigma_points - mean) * (lik_expectation - mu), axis=-1
        w * quadratic_part2.T, axis=-1
    ).T
    # compute equivalent likelihood noise, omega
    omega = S - C.T @ solve(cov, C)
    return np.squeeze(mu), omega


def expected_conditional_mean_dm(likelihood, mean, cov, cubature=None):
    """
    """
    dmu_dm = jacrev(expected_conditional_mean_cubature, argnums=1)(likelihood, mean, cov, cubature)[0]
    return np.squeeze(dmu_dm)


def expected_conditional_mean_dm2(likelihood, mean, cov, cubature=None):
    """
    """
    d2mu_dm2 = jacrev(expected_conditional_mean_dm, argnums=1)(likelihood, mean, cov, cubature)
    return np.squeeze(d2mu_dm2, axis=-1)


def predict_cubature(likelihood, mean_f, var_f, cubature=None):
    """
    predict in data space given predictive mean and var of the latent function
    """
    if cubature is None:
        x, w = gauss_hermite(mean_f.shape[0])  # Gauss-Hermite sigma points and weights
    else:
        x, w = cubature(mean_f.shape[0])
    var_f = (var_f + var_f.T) / 2
    chol_f, low = cho_factor(var_f, lower=True)
    # fsigᵢ=xᵢ√cₙ + mₙ: scale locations according to latent dist.
    sigma_points = chol_f @ np.atleast_2d(x) + mean_f
    # Compute moments via cubature:
    # E[y] = ∫ E[yₙ|fₙ] N(fₙ|mₙ,vₙ) dfₙ
    #      ≈ ∑ᵢ wᵢ E[yₙ|fₙ]
    # E[y^2] = ∫ (Cov[yₙ|fₙ] + E[yₙ|fₙ]^2) N(fₙ|mₙ,vₙ) dfₙ
    #        ≈ ∑ᵢ wᵢ (Cov[yₙ|fₙ] + E[yₙ|fₙ]^2)
    conditional_expectation, conditional_covariance = likelihood.conditional_moments(sigma_points)
    expected_y = np.sum(w * conditional_expectation, axis=-1)
    conditional_expectation_ = conditional_expectation.T[..., None]
    conditional_expectation_squared = conditional_expectation_ @ transpose(conditional_expectation_)
    expected_y_squared = np.sum(
        w * (conditional_covariance + conditional_expectation_squared.T),
        axis=-1
    )
    # Cov[y] = E[y^2] - E[y]^2
    covariance_y = expected_y_squared - expected_y[..., None] @ expected_y[None]
    return expected_y, covariance_y



import jax.numpy as np
from jax import vmap
from jax.ops import index_add, index_update, index
from jax.scipy.linalg import cho_factor, cho_solve
from jax.scipy.linalg import solve as jsc_solve
from .utils import mvn_logpdf, solve, transpose, inv, inv_vmap
from jax.lax import scan, associative_scan
import math

INV2PI = (2 * math.pi) ** -1


def get_diag_and_offdiag_components(num_latents, zeros, i, noise_cov):
    temp_vec = index_add(zeros, index[i], 1.)
    temp_mat = temp_vec.reshape(num_latents, num_latents)
    return np.kron(np.diag(noise_cov), temp_mat)  # block-diag


def blocktensor_to_blockdiagmatrix(blocktensor):
    """
    Convert [N, D, D] tensor to [ND, ND] block-diagonal matrix
    """
    N = blocktensor.shape[0]
    D = blocktensor.shape[1]
    diag_and_offdiag_components = vmap(get_diag_and_offdiag_components, in_axes=(None, None, 0, 1))(
        D, np.zeros(D ** 2), np.arange(D ** 2), blocktensor.reshape(N, -1)
    )
    return np.sum(diag_and_offdiag_components, axis=0)


def get_blocks(blockdiagmatrix, D, i):
    return blockdiagmatrix[0+D*i:D+D*i, 0+D*i:D+D*i]


@vmap
def get_3d_off_diag(offdiag_elems):
    return np.sum(
        offdiag_elems[:, None, None]
        * np.array([[[0., 1., 0.], [1., 0., 0.], [0., 0., 0.]],
                    [[0., 0., 0.], [0., 0., 1.], [0., 1., 0.]]]),
        axis=0
    )


def blockdiagmatrix_to_blocktensor(blockdiagmatrix, N, D):
    """
    Convert [ND, ND] block-diagonal matrix to [N, D, D] tensor
    TODO: extend to D>3 case
    """
    diags = vmap(np.diag)(np.diag(blockdiagmatrix).reshape(N, D))
    if D == 1:
        blocktensors = diags
    elif D == 2:
        offdiag_elems = np.diag(np.concatenate([np.zeros([N * D, 1]), blockdiagmatrix], axis=1)).reshape(N, D)[:, 1:]
        offdiags = offdiag_elems[..., None] * np.fliplr(np.eye(D))
        blocktensors = diags + offdiags
    elif D == 3:
        addzeros = np.concatenate([np.zeros([N * D, 1]), blockdiagmatrix], axis=1)
        addzeros2 = np.concatenate([np.zeros([N * D, 1]), addzeros], axis=1)
        offdiag_elems = np.diag(addzeros).reshape(N, D)[:, 1:]
        offdiags = get_3d_off_diag(offdiag_elems)
        corner_elements = np.diag(addzeros2).reshape(N, D)[:, 2:]
        corners = corner_elements[..., None] * np.array([[0., 0., 1.], [0., 0., 0.], [1., 0., 0.]])
        blocktensors = diags + offdiags + corners
    else:
        raise NotImplementedError('Multi-latent case with D>3 not implemented')
    return blocktensors


def gaussian_conditional(kernel, y, noise_cov, X, X_star=None):
    """
    Compute the GP posterior / predictive distribution using standard Gaussian identities
    :param kernel: an instantiation of the kernel class
    :param y: observations [N, 1]
    :param noise_cov: observation noise covariance [N, 1]
    :param X: training inputs [N, D]
    :param X_star: test inputs [N*, D]
    :return:
        mean: posterior mean [N, 1]
        covariance: posterior covariance [N, N]
    """
    Kff = kernel(X, X)
    if X_star is None:  # inference / learning
        Kfs = Kff
        Kss = Kff
    else:  # prediction
        Kfs = kernel(X, X_star)
        Kss = kernel(X_star, X_star)

    # Ky = Kff + np.diag(np.squeeze(noise_cov))  # single-latent version
    noise_cov_block_diag = blocktensor_to_blockdiagmatrix(noise_cov)  # multi-latent version
    Ky = Kff + noise_cov_block_diag
    # ---- compute approximate posterior using standard Gaussian conditional formula ----
    Kfs_iKy = solve(Ky, Kfs).T
    # mean = Kfs_iKy @ diag(y)
    mean = Kfs_iKy @ y.reshape(-1, 1)
    covariance = Kss - Kfs_iKy @ Kfs
    return mean, covariance


def sparse_gaussian_conditional(kernel, nat1lik, nat2lik, X, Z):
    """
    Compute q(u)
    :param kernel: an instantiation of the kernel class
    :param nat1lik: likelihood first natural parameter [N, 1]
    :param nat2lik: likelihood noise precision [N, 1]
    :param X: training inputs [N, D]
    :param Z: inducing inputs [N*, D]
    :return:
        mean: posterior mean [N, 1]
        covariance: posterior covariance [N, N]
    """
    Kuf = kernel(Z, X)
    Kuu = kernel(Z, Z)
    # nat2prior = inv(Kuu)
    Wuf = solve(Kuu, Kuf)  # conditional mapping, Kuu^-1 Kuf

    nat1lik_fullrank = Wuf @ nat1lik.reshape(-1, 1)
    # nat2lik_fullrank = Wuf @ np.diag(np.squeeze(nat2lik)) @ transpose(Wuf)
    nat2lik_block_diag = blocktensor_to_blockdiagmatrix(nat2lik)  # multi-latent version
    nat2lik_fullrank = Wuf @ nat2lik_block_diag @ transpose(Wuf)

    # nat1post = nat1lik_fullrank  # prior nat1 is zero
    # nat2post = nat2prior + nat2lik_fullrank

    # covariance = inv(nat2post)
    # mean = covariance @ nat1post

    likcov = inv(nat2lik_fullrank)
    likmean = likcov @ nat1lik_fullrank
    Ky = Kuu + likcov
    # ---- compute approximate posterior using standard Gaussian conditional formula ----
    Kuu_iKy = solve(Ky, Kuu).T
    mean = Kuu_iKy @ likmean
    covariance = Kuu - Kuu_iKy @ Kuu

    return mean, covariance


def sparse_conditional_post_to_data(kernel, post_mean, post_cov, X, Z):
    """
    Compute int p(f|u) q(u) du
    :param kernel: an instantiation of the kernel class
    :param post_mean: posterior mean [M, 1]
    :param post_cov: posterior covariance [M, M]
    :param X: training inputs [N, D]
    :param Z: inducing inputs [N*, D]
    :return:
        mean: posterior mean [N, 1]
        covariance: posterior covariance [N, N]
    """
    X = X.reshape(X.shape[0], -1)
    Kff = kernel(X, X)
    Kuf = kernel(Z, X)
    Kuu = kernel(Z, Z)
    Wuf = solve(Kuu, Kuf)  # conditional mapping, Kuu^-1 Kuf
    Qff = transpose(Kuf) @ Wuf  # Kfu Kuu^-1 Kuf

    conditional_cov = Kff - Qff

    mean_f = transpose(Wuf) @ post_mean.reshape(-1, 1)
    cov_f = conditional_cov + transpose(Wuf) @ post_cov @ Wuf

    return mean_f, cov_f


def process_noise_covariance(A, Pinf):
    Q = Pinf - A @ Pinf @ transpose(A)
    return Q


def _sequential_kf(As, Qs, H, ys, noise_covs, m0, P0, masks, return_predict=False):

    def body(carry, inputs):
        y, A, Q, obs_cov, mask = inputs
        m, P, ell = carry
        m_ = A @ m
        P_ = A @ P @ A.T + Q

        obs_mean = H @ m_
        HP = H @ P_
        S = HP @ H.T + obs_cov

        ell_n = mvn_logpdf(y, obs_mean, S, mask)
        ell = ell + ell_n

        K = solve(S, HP).T
        m = m_ + K @ (y - obs_mean)
        P = P_ - K @ HP
        if return_predict:
            return (m, P, ell), (m_, P_)
        else:
            return (m, P, ell), (m, P)

    (_, _, loglik), (fms, fPs) = scan(f=body,
                                      init=(m0, P0, 0.),
                                      xs=(ys, As, Qs, noise_covs, masks))
    return loglik, fms, fPs


def parallel_filtering_element_(A, Q, H, noise_cov, y):
    HQ, HA = H @ Q, H @ A  # pre-compute intermediates
    S = HQ @ H.T + noise_cov  # H Q H.T + R

    SinvH = solve(S, H)  # S^{-1} H

    K = Q @ SinvH.T  # Q H.T S^{-1}
    AA = A - K @ HA  # A - K H A
    b = K @ y  # K y
    C = Q - K @ HQ  # Q - K H Q

    SinvHA = (SinvH @ A).T  # A.T H.T S^{-1}
    eta = SinvHA @ y  # A.T H.T S^{-1} y
    J = SinvHA @ HA  # A.T H.T S^{-1} H A
    return AA, b, C, J, eta


parallel_filtering_element = vmap(parallel_filtering_element_, in_axes=(0, 0, None, 0, 0))


@vmap
def parallel_filtering_operator(elem1, elem2):
    A1, b1, C1, J1, eta1 = elem1
    A2, b2, C2, J2, eta2 = elem2

    C1inv = inv(C1)
    temp = solve(C1inv + J2, C1inv)  # we should avoid inverting non-PSD matrices here
    A2temp = A2 @ temp
    AA = A2temp @ A1
    b = A2temp @ (b1 + C1 @ eta2) + b2
    C = A2temp @ C1 @ A2.T + C2

    A1temp = A1.T @ temp.T  # re-use previous solve
    eta = A1temp @ (eta2 - J2 @ b1) + eta1
    J = A1temp @ J2 @ A1 + J1

    return AA, b, C, J, eta


def make_associative_filtering_elements(As, Qs, H, ys, noise_covs, m0, P0):
    Qs = index_update(Qs, index[0], P0)  # first element requires different initialisation
    AA, b, C, J, eta = parallel_filtering_element(As, Qs, H, noise_covs, ys)
    # modify initial b to account for m0 (not needed if m0=zeros)
    S = H @ Qs[0] @ H.T + noise_covs[0]
    K0 = solve(S, H @ Qs[0]).T
    b = index_add(b, index[0], m0 - K0 @ H @ m0)
    return AA, b, C, J, eta


@vmap
def vmap_mvn_logpdf(*args, **kwargs):
    return mvn_logpdf(*args, **kwargs)


def _parallel_kf(As, Qs, H, ys, noise_covs, m0, P0, masks, return_predict=False):

    # perform parallel filtering
    initial_elements = make_associative_filtering_elements(As, Qs, H, ys, noise_covs, m0, P0)
    final_elements = associative_scan(parallel_filtering_operator, initial_elements)
    fms, fPs = final_elements[1], final_elements[2]

    # now compute the log likelihood
    mpredict = As @ np.concatenate([m0[None], fms[:-1]])
    Ppredict = As @ np.concatenate([P0[None], fPs[:-1]]) @ transpose(As) + Qs

    loglik = np.sum(vmap_mvn_logpdf(ys, H @ mpredict, H @ Ppredict @ H.T + noise_covs, masks))

    if return_predict:
        return loglik, mpredict, Ppredict
    else:
        return loglik, fms, fPs


def kalman_filter(dt, kernel, y, noise_cov, mask=None, parallel=False, return_predict=False):
    """
    Run the Kalman filter to get p(fₙ|y₁,...,yₙ).
    Assumes a heteroscedastic Gaussian observation model, i.e. var is vector valued
    :param dt: step sizes [N, 1]
    :param kernel: an instantiation of the kernel class, used to determine the state space model
    :param y: observations [N, D, 1]
    :param noise_cov: observation noise covariances [N, D, D]
    :param mask: boolean mask for the observations (to indicate missing data locations) [N, D, 1]
    :param parallel: flag to switch between parallel and sequential implementation of Kalman filter
    :param return_predict: flag whether to return predicted state, rather than updated state
    :return:
        ell: the log-marginal likelihood log p(y), for hyperparameter optimisation (learning) [scalar]
        means: intermediate filtering means [N, state_dim, 1]
        covs: intermediate filtering covariances [N, state_dim, state_dim]
    """
    if mask is None:
        mask = np.zeros_like(y, dtype=bool)
    Pinf = kernel.stationary_covariance()
    minf = np.zeros([Pinf.shape[0], 1])

    As = vmap(kernel.state_transition)(dt)
    Qs = vmap(process_noise_covariance, [0, None])(As, Pinf)
    H = kernel.measurement_model()

    if parallel:
        ell, means, covs = _parallel_kf(As, Qs, H, y, noise_cov, minf, Pinf, mask, return_predict=return_predict)
    else:
        ell, means, covs = _sequential_kf(As, Qs, H, y, noise_cov, minf, Pinf, mask, return_predict=return_predict)
    return ell, (means, covs)


def _sequential_rts(fms, fPs, As, Qs, H, return_full):

    def body(carry, inputs):
        fm, fP, A, Q = inputs
        sm, sP = carry

        pm = A @ fm
        AfP = A @ fP
        pP = AfP @ A.T + Q

        C = solve(pP, AfP).T

        sm = fm + C @ (sm - pm)
        sP = fP + C @ (sP - pP) @ C.T
        if return_full:
            return (sm, sP), (sm, sP, C)
        else:
            return (sm, sP), (H @ sm, H @ sP @ H.T, C)

    _, (sms, sPs, gains) = scan(f=body,
                                init=(fms[-1], fPs[-1]),
                                xs=(fms, fPs, As, Qs),
                                reverse=True)
    return sms, sPs, gains


def last_parallel_smoothing_element(m, P):
    return np.zeros_like(P), m, P


@vmap
def parallel_smoothing_element(A, Q, m, P):
    Pp = A @ P @ transpose(A) + Q

    E = transpose(solve(Pp, A @ P))
    g = m - E @ A @ m
    L = P - E @ Pp @ transpose(E)
    return E, g, L


@vmap
def parallel_smoothing_operator(elem1, elem2):
    E1, g1, L1 = elem1
    E2, g2, L2 = elem2
    E = E2 @ E1
    g = E2 @ g1 + g2
    L = E2 @ L1 @ transpose(E2) + L2
    return E, g, L


def _parallel_rts(fms, fPs, As, Qs, H, return_full):

    # build the associative smoothing elements
    smoothing_elems = parallel_smoothing_element(As, Qs, fms, fPs)
    gains = smoothing_elems[0]
    last_elems = last_parallel_smoothing_element(fms[-1], fPs[-1])
    initial_elements = tuple(np.append(gen_es[:-1], np.expand_dims(last_e, 0), axis=0)
                             for gen_es, last_e in zip(smoothing_elems, last_elems))

    # run the parallel smoother
    final_elements = associative_scan(parallel_smoothing_operator, initial_elements, reverse=True)  # note vmap
    sms, sPs = final_elements[1], final_elements[2]

    if return_full:
        return sms, sPs, gains
    else:
        return H @ sms, H @ sPs @ H.T, gains


def rauch_tung_striebel_smoother(dt, kernel, filter_mean, filter_cov, return_full=False, parallel=False):
    """
    Run the RTS smoother to get p(fₙ|y₁,...,y_N),
    :param dt: step sizes [N, 1]
    :param kernel: an instantiation of the kernel class, used to determine the state space model
    :param filter_mean: the intermediate distribution means computed during filtering [N, state_dim, 1]
    :param filter_cov: the intermediate distribution covariances computed during filtering [N, state_dim, state_dim]
    :param return_full: a flag determining whether to return the full state distribution or just the function(s)
    :param parallel: flag to switch between parallel and sequential implementation of smoother
    :return:
        smoothed_mean: the posterior marginal means [N, obs_dim]
        smoothed_var: the posterior marginal variances [N, obs_dim]
    """
    Pinf = kernel.stationary_covariance()

    As = vmap(kernel.state_transition)(dt)
    Qs = vmap(process_noise_covariance, [0, None])(As, Pinf)
    H = kernel.measurement_model()

    if parallel:
        means, covs, gains = _parallel_rts(filter_mean, filter_cov, As, Qs, H, return_full)
    else:
        means, covs, gains = _sequential_rts(filter_mean, filter_cov, As, Qs, H, return_full)
    return means, covs, gains


def kalman_filter_pairs(dt, kernel, y, noise_cov, mask=None, parallel=False):
    """
    A Kalman filter over pairs of states, in which y is [2state_dim, 1] and noise_cov is [2state_dim, 2state_dim]
    :param dt: step sizes [N, 1]
    :param kernel: an instantiation of the kernel class, used to determine the state space model
    :param y: observations [N, 2state_dim, 1]
    :param noise_cov: observation noise covariances [N, 2state_dim, 2state_dim]
    :param mask: boolean mask for the observations (to indicate missing data locations) [N, 2state_dim, 1]
    :param parallel: flag to switch between parallel and sequential implementation of Kalman filter
    :return:
        ell: the log-marginal likelihood log p(y), for hyperparameter optimisation (learning) [scalar]
        means: marginal state filtering means [N, state_dim, 1]
        covs: marginal state filtering covariances [N, state_dim, state_dim]
    """
    if mask is None:
        mask = np.zeros_like(y, dtype=bool)

    Pinf = kernel.stationary_covariance()
    state_dim = Pinf.shape[0]
    minf = np.zeros([state_dim, 1])
    zeros = np.zeros([state_dim, state_dim])
    Pinfpair = np.block([[Pinf,  zeros],
                         [zeros, Pinf]])
    minfpair = np.block([[minf],
                         [minf]])

    As = vmap(kernel.state_transition)(dt)
    Qs = vmap(process_noise_covariance, [0, None])(As, Pinf)

    def construct_pair(A, Q):
        Apair = np.block([[zeros, np.eye(state_dim)],
                          [zeros, A]])
        Qpair = np.block([[1e-32 * np.eye(state_dim), zeros],  # jitter avoids numerical errors in parallel filter init
                          [zeros,                     Q]])
        return Apair, Qpair

    Apairs, Qpairs = vmap(construct_pair)(As, Qs)
    H = np.eye(2 * state_dim)

    if parallel:
        ell, means, covs = _parallel_kf(Apairs, Qpairs, H, y, noise_cov, minfpair, Pinfpair, mask)
    else:
        ell, means, covs = _sequential_kf(Apairs, Qpairs, H, y, noise_cov, minfpair, Pinfpair, mask)
    return ell, (means[1:, :state_dim], covs[1:, :state_dim, :state_dim])


def _sequential_kf_mf(As, Qs, H, ys, noise_covs, m0, P0, masks, block_index):

    def build_block_diag(P_blocks):
        P = index_add(Pzeros, index[block_index], P_blocks.flatten())
        return P

    def get_block_mean(m):
        return m.reshape(num_latents, sub_state_dim, 1)

    def get_block_cov(P):
        return P[block_index].reshape(num_latents, sub_state_dim, sub_state_dim)

    num_latents = m0.shape[0]
    sub_state_dim = m0.shape[1]
    state_dim = num_latents * sub_state_dim
    Pzeros = np.zeros([state_dim, state_dim])

    def body(carry, inputs):
        y, A, Q, obs_cov, mask = inputs
        m, P, ell = carry
        m = (A @ m).reshape(-1, 1)
        P = build_block_diag(A @ P @ transpose(A) + Q)

        obs_mean = H @ m
        HP = H @ P
        S = HP @ H.T + obs_cov

        ell_n = mvn_logpdf(y, obs_mean, S, mask)
        ell = ell + ell_n

        K = solve(S, HP).T
        m = get_block_mean(m + K @ (y - obs_mean))
        P = get_block_cov(P - K @ HP)
        return (m, P, ell), (m, P)

    (_, _, loglik), (fms, fPs) = scan(f=body,
                                      init=(m0, P0, 0.),
                                      xs=(ys, As, Qs, noise_covs, masks))
    return loglik, fms, fPs


def parallel_filtering_element_mf_(A, Q, H, noise_cov, y, block_index):

    def build_block_diag(P_blocks):
        P = index_add(Pzeros, index[block_index], P_blocks.flatten())
        return P

    def get_block_mean(m):
        return m.reshape(num_latents, sub_state_dim, 1)

    def get_block_cov(P):
        return P[block_index].reshape(num_latents, sub_state_dim, sub_state_dim)

    num_latents = A.shape[0]
    sub_state_dim = A.shape[1]
    state_dim = num_latents * sub_state_dim
    Pzeros = np.zeros([state_dim, state_dim])

    Q, A = build_block_diag(Q), build_block_diag(A)
    HQ, HA = H @ Q, H @ A  # pre-compute intermediates
    S = HQ @ H.T + noise_cov  # H Q H.T + R

    SinvH = solve(S, H)  # S^{-1} H

    K = Q @ SinvH.T  # Q H.T S^{-1}
    AA = get_block_cov(A - K @ HA)  # A - K H A
    b = get_block_mean(K @ y)  # K y
    C = get_block_cov(Q - K @ HQ)  # Q - K H Q

    SinvHA = (SinvH @ A).T  # A.T H.T S^{-1}
    eta = get_block_mean(SinvHA @ y)  # A.T H.T S^{-1} y
    J = get_block_cov(SinvHA @ HA)  # A.T H.T S^{-1} H A
    return AA, b, C, J, eta


parallel_filtering_element_mf = vmap(parallel_filtering_element_mf_, in_axes=(0, 0, None, 0, 0, None))


@vmap
def parallel_filtering_operator_mf(elem1, elem2):
    A1, b1, C1, J1, eta1 = elem1
    A2, b2, C2, J2, eta2 = elem2

    C1inv = inv_vmap(C1)
    temp = solve(C1inv + J2, C1inv)  # we should avoid inverting non-PSD matrices here
    A2temp = A2 @ temp
    AA = A2temp @ A1
    b = A2temp @ (b1 + C1 @ eta2) + b2
    C = A2temp @ C1 @ transpose(A2) + C2

    A1temp = transpose(A1) @ transpose(temp)  # re-use previous solve
    eta = A1temp @ (eta2 - J2 @ b1) + eta1
    J = A1temp @ J2 @ A1 + J1

    return AA, b, C, J, eta


def make_associative_filtering_elements_mf(As, Qs, H, ys, noise_covs, m0, P0, block_index):

    def build_block_diag(P_blocks):
        P = index_add(Pzeros, index[block_index], P_blocks.flatten())
        return P

    def get_block_mean(m):
        return m.reshape(num_latents, sub_state_dim, 1)

    num_latents = As.shape[1]
    sub_state_dim = As.shape[2]
    state_dim = num_latents * sub_state_dim
    Pzeros = np.zeros([state_dim, state_dim])

    Qs = index_update(Qs, index[0], P0)  # first element requires different initialisation
    AA, b, C, J, eta = parallel_filtering_element_mf(As, Qs, H, noise_covs, ys, block_index)
    # modify initial b to account for m0 (not needed if m0=zeros)
    m0 = m0.reshape(-1, 1)
    Qs0 = build_block_diag(Qs[0])
    S = H @ Qs0 @ H.T + noise_covs[0]
    K0 = solve(S, H @ Qs0).T
    b = index_add(b, index[0], get_block_mean(m0 - K0 @ (H @ m0)))
    return AA, b, C, J, eta


def _parallel_kf_mf(As, Qs, H, ys, noise_covs, m0, P0, masks, block_index):

    @vmap
    def build_block_diag(P_blocks):
        P = index_add(Pzeros, index[block_index], P_blocks.flatten())
        return P

    def build_mean(m):
        return m.reshape(num_time_steps, state_dim, 1)

    num_time_steps = As.shape[0]
    num_latents = As.shape[1]
    sub_state_dim = As.shape[2]
    state_dim = num_latents * sub_state_dim
    Pzeros = np.zeros([state_dim, state_dim])

    # perform parallel filtering
    initial_elements = make_associative_filtering_elements_mf(As, Qs, H, ys, noise_covs, m0, P0, block_index)
    final_elements = associative_scan(parallel_filtering_operator_mf, initial_elements)
    fms, fPs = final_elements[1], final_elements[2]

    # now compute the log likelihood
    mpredict = build_mean(As @ np.concatenate([m0[None], fms[:-1]]))
    Ppredict = build_block_diag(As @ np.concatenate([P0[None], fPs[:-1]]) @ transpose(As) + Qs)

    loglik = np.sum(vmap_mvn_logpdf(ys, H @ mpredict, H @ Ppredict @ H.T + noise_covs, masks))

    return loglik, fms, fPs


def kalman_filter_meanfield(dt, kernel, y, noise_cov, mask=None, parallel=False, block_index=None):
    """
    Run the Kalman filter to get p(fₙ|y₁,...,yₙ) with a mean-field approximation for the latent components.
    Assumes a heteroscedastic Gaussian observation model, i.e. var is vector valued
    :param dt: step sizes [N, 1]
    :param kernel: an instantiation of the kernel class, used to determine the state space model
    :param y: observations [N, D, 1]
    :param noise_cov: observation noise covariances [N, D, D]
    :param mask: boolean mask for the observations (to indicate missing data locations) [N, D, 1]
    :param parallel: flag to switch between parallel and sequential implementation of Kalman filter
    :param block_index: mean-field block indices required to build the block-diagonal model matrices from the blocks
    :return:
        ell: the log-marginal likelihood log p(y), for hyperparameter optimisation (learning) [scalar]
        means: intermediate filtering means [N, state_dim, 1]
        covs: intermediate filtering covariances [N, state_dim, state_dim]
    """
    if mask is None:
        mask = np.zeros_like(y, dtype=bool)
    Pinf = kernel.stationary_covariance_meanfield()
    minf = np.zeros([Pinf.shape[0], Pinf.shape[1], 1])

    As = vmap(kernel.state_transition_meanfield)(dt)
    Qs = vmap(process_noise_covariance, [0, None])(As, Pinf)
    H = kernel.measurement_model()

    if parallel:
        ell, means, covs = _parallel_kf_mf(As, Qs, H, y, noise_cov, minf, Pinf, mask, block_index)
    else:
        ell, means, covs = _sequential_kf_mf(As, Qs, H, y, noise_cov, minf, Pinf, mask, block_index)
    return ell, (means, covs)


def _sequential_rts_mf(fms, fPs, As, Qs, H, return_full, block_index):

    def build_block_diag(P_blocks):
        P = index_add(Pzeros, index[block_index], P_blocks.flatten())
        return P

    num_latents = fms.shape[1]
    sub_state_dim = fms.shape[2]
    state_dim = num_latents * sub_state_dim
    Pzeros = np.zeros([state_dim, state_dim])

    def body(carry, inputs):
        fm, fP, A, Q = inputs
        sm, sP = carry

        pm = A @ fm
        AfP = A @ fP
        pP = AfP @ transpose(A) + Q

        C = transpose(solve(pP, AfP))

        sm = fm + C @ (sm - pm)
        sP = fP + C @ (sP - pP) @ transpose(C)
        if return_full:
            return (sm, sP), (sm.reshape(-1, 1), build_block_diag(sP), build_block_diag(C))
        else:
            return (sm, sP), (H @ sm.reshape(-1, 1), H @ build_block_diag(sP) @ transpose(H), build_block_diag(C))

    _, (sms, sPs, gains) = scan(f=body,
                                init=(fms[-1], fPs[-1]),
                                xs=(fms, fPs, As, Qs),
                                reverse=True)
    return sms, sPs, gains


def _parallel_rts_mf(fms, fPs, As, Qs, H, return_full, block_index):

    @vmap
    def build_block_diag(P_blocks):
        P = index_add(Pzeros, index[block_index], P_blocks.flatten())
        return P

    def build_mean(m):
        return m.reshape(num_time_steps, state_dim, 1)

    num_time_steps = As.shape[0]
    num_latents = As.shape[1]
    sub_state_dim = As.shape[2]
    state_dim = num_latents * sub_state_dim
    Pzeros = np.zeros([state_dim, state_dim])

    # build the associative smoothing elements
    smoothing_elems = parallel_smoothing_element(As, Qs, fms, fPs)
    gains = build_block_diag(smoothing_elems[0])
    last_elems = last_parallel_smoothing_element(fms[-1], fPs[-1])
    initial_elements = tuple(np.append(gen_es[:-1], np.expand_dims(last_e, 0), axis=0)
                             for gen_es, last_e in zip(smoothing_elems, last_elems))

    # run the parallel smoother
    final_elements = associative_scan(parallel_smoothing_operator, initial_elements, reverse=True)  # note vmap
    sms, sPs = build_mean(final_elements[1]), build_block_diag(final_elements[2])

    if return_full:
        return sms, sPs, gains
    else:
        return H @ sms, H @ sPs @ H.T, gains


def rauch_tung_striebel_smoother_meanfield(dt, kernel, filter_mean, filter_cov, return_full=False,
                                           parallel=False, block_index=None):
    """
    Run the RTS smoother to get p(fₙ|y₁,...,y_N) with a mean-field approximation for the latent components.
    :param dt: step sizes [N, 1]
    :param kernel: an instantiation of the kernel class, used to determine the state space model
    :param filter_mean: the intermediate distribution means computed during filtering [N, state_dim, 1]
    :param filter_cov: the intermediate distribution covariances computed during filtering [N, state_dim, state_dim]
    :param return_full: a flag determining whether to return the full state distribution or just the function(s)
    :param parallel: flag to switch between parallel and sequential implementation of smoother
    :param block_index: mean-field block indiced required to build the block-diagonal model matrices from the blocks
    :return:
        smoothed_mean: the posterior marginal means [N, obs_dim]
        smoothed_var: the posterior marginal variances [N, obs_dim]
    """
    Pinf = kernel.stationary_covariance_meanfield()

    As = vmap(kernel.state_transition_meanfield)(dt)
    Qs = vmap(process_noise_covariance, [0, None])(As, Pinf)
    H = kernel.measurement_model()

    if parallel:
        means, covs, gains = _parallel_rts_mf(filter_mean, filter_cov, As, Qs, H, return_full, block_index)
    else:
        means, covs, gains = _sequential_rts_mf(filter_mean, filter_cov, As, Qs, H, return_full, block_index)
    return means, covs, gains


def _sequential_kf_pairs_mf(As, Qs, ys, noise_covs, m0, P0, block_index):

    def build_block_diag(P_blocks):
        P = np.zeros([state_dim, state_dim])
        P = index_add(P, index[block_index], P_blocks.flatten())
        return P

    def get_block_mean(m):
        return m.reshape(num_latents, sub_state_dim, 1)

    def get_block_cov(P):
        return P[block_index].reshape(num_latents, sub_state_dim, sub_state_dim)

    num_latents = m0.shape[0]
    sub_state_dim = m0.shape[1]
    state_dim = num_latents * sub_state_dim

    def body(carry, inputs):
        y, A, Q, obs_cov = inputs
        m_left, P_left, ell = carry

        # predict
        m_right = (A @ m_left).reshape(-1, 1)
        P_right = build_block_diag(A @ P_left @ transpose(A) + Q)

        # construct the joint distribution p(uₙ₋₁,uₙ) = p(uₙ₋₁)p(uₙ|uₙ₋₁)
        PA_ = build_block_diag(P_left @ transpose(A))
        m_joint = np.block([[m_left.reshape(-1, 1)],
                            [m_right]])
        P_joint = np.block([[build_block_diag(P_left), PA_],
                            [PA_.T, P_right]])

        S = P_joint + obs_cov

        ell_n = mvn_logpdf(y, m_joint, S)
        ell = ell + ell_n

        K = solve(S, P_joint).T

        # perform update
        m = m_joint + K @ (y - m_joint)
        P = P_joint - P_joint @ K.T

        # marginalise and store the now fully updated left state, uₙ₋₁
        m_left = get_block_mean(m[:state_dim])
        P_left = get_block_cov(P[:state_dim, :state_dim])
        # marginalise and propagate the right state, uₙ
        m_right = get_block_mean(m[state_dim:])
        P_right = get_block_cov(P[state_dim:, state_dim:])

        return (m_right, P_right, ell), (m_left, P_left)

    (_, _, loglik), (fms, fPs) = scan(f=body,
                                      init=(m0, P0, 0.),
                                      xs=(ys, As, Qs, noise_covs))
    return loglik, fms[1:], fPs[1:]  # discard intial dummy state


def kalman_filter_pairs_meanfield(dt, kernel, y, noise_cov, parallel=False, block_index=None):
    """
    A Kalman filter over pairs of states, with a mean-field approximation for the latent components.
    Assumes a heteroscedastic Gaussian observation model, i.e. var is vector valued
    :param dt: step sizes [N, 1]
    :param kernel: an instantiation of the kernel class, used to determine the state space model
    :param y: observations [N, 2state_dim, 1]
    :param noise_cov: observation noise covariances [N, 2state_dim, 2state_dim]
    :param parallel: flag to switch between parallel and sequential implementation of Kalman filter
    :param block_index: mean-field block indices required to build the block-diagonal model matrices from the blocks
    :return:
        ell: the log-marginal likelihood log p(y), for hyperparameter optimisation (learning) [scalar]
        means: intermediate filtering means [N, state_dim, 1]
        covs: intermediate filtering covariances [N, state_dim, state_dim]
    """

    Pinf = kernel.stationary_covariance_meanfield()
    minf = np.zeros([Pinf.shape[0], Pinf.shape[1], 1])

    As = vmap(kernel.state_transition_meanfield)(dt)
    Qs = vmap(process_noise_covariance, [0, None])(As, Pinf)

    if parallel:
        raise NotImplementedError("Parallel KF not implemented yet")
    else:
        ell, means, covs = _sequential_kf_pairs_mf(As, Qs, y, noise_cov, minf, Pinf, block_index)
    return ell, (means, covs)


def dare(A, H, Q, R, Pinit, num_iters=20):
    """
    solve a discrete algebraic Ricatti equation
    :param A: transition matrix
    :param H: measurement model mean
    :param Q: process noise covariance
    :param R: observation noise covariance
    :param Pinit: initialisation
    :param num_iters: number of iterations
    :return:
    """

    # def body(X, _):
    #     K = A @ transpose(solve(H @ X @ transpose(H) + R, H @ X))
    #     X = (A - K @ H) @ X @ transpose(A - K @ H) + K @ R @ transpose(K) + Q
    #     return X, X

    def body(X, _):  # slightly faster version
        HX = H @ X
        S = HX @ H.T + R
        K = solve(S, HX).T
        X = A @ (X - K @ HX) @ A.T + Q
        return X, X

    X_final, _ = scan(f=body,
                      init=Pinit,
                      xs=np.zeros(num_iters))

    return X_final


def _sequential_kf_ih(AKHAs, HA, Kys, Ss, ys, m0, masks):

    def body(carry, inputs):
        y, AKHA, Ky, S, mask = inputs
        m, ell = carry

        obs_mean = HA @ m

        ell_n = mvn_logpdf(y, obs_mean, S, mask)
        ell = ell + ell_n

        # perform update
        m = AKHA @ m + Ky

        return (m, ell), m

    (_, loglik), (fms) = scan(f=body,
                              init=(m0, 0.),
                              xs=(ys, AKHAs, Kys, Ss, masks))
    return loglik, fms


@vmap
def parallel_filtering_operator_ih_homoscedastic(elem1, elem2):
    B1, v1 = elem1
    B2, v2 = elem2
    return B1 @ B2, B2 @ v1 + v2


@vmap
def parallel_filtering_operator_ih_heteroscedastic(elem1, elem2):
    B1, v1 = elem1
    B2, v2 = elem2
    return B1 @ B2, jsc_solve(B2, v1, sym_pos=False) + v2


def _parallel_kf_ih(AKHAs_or_iAKHAs, HA, Kys, Ss, ys, m0, masks, heteroscedastic):

    # perform parallel filtering
    if heteroscedastic:
        # Kys = index_add(Kys, index[0], AKHAs[1] @ m0)  # incorporate m0, TODO: not sure if correct, check
        final_elements = associative_scan(parallel_filtering_operator_ih_heteroscedastic, (AKHAs_or_iAKHAs, Kys))
    else:
        # Kys = index_add(Kys, index[0], jsc_inv(AKHAs[1]) @ m0)  # incorporate m0, TODO: not sure if correct, check
        final_elements = associative_scan(parallel_filtering_operator_ih_homoscedastic, (AKHAs_or_iAKHAs, Kys))
    fms = final_elements[1]

    # now compute the log likelihood
    obs_means = HA @ np.concatenate([m0[None], fms[:-1]])
    loglik = np.sum(vmap_mvn_logpdf(ys, obs_means, Ss, masks))

    return loglik, fms


def kalman_filter_infinite_horizon(dt, kernel, y, noise_cov, mask=None, parallel=False,
                                   heteroscedastic=False, noise_cov_tied=None, dare_iters=20, dare_init=None):
    """
    Run the Kalman filter to get p(fₙ|y₁,...,yₙ).
    Assumes a heteroscedastic Gaussian observation model, i.e. var is vector valued
    :param dt: step sizes [N, 1]
    :param kernel: an instantiation of the kernel class, used to determine the state space model
    :param y: observations [N, D, 1]
    :param noise_cov: observation noise covariances [N, D, D]
    :param mask: boolean mask for the observations (to indicate missing data locations) [N, D, 1]
    :param parallel: flag to switch between parallel and sequential implementation of Kalman filter
    :param heteroscedastic: flag to indicate whether the approximate model has heteroscedastic noise
    :param noise_cov_tied: averaged observation noise covariance [N, D, D]
    :param dare_iters: number of iterations to run the DARE solver for
    :param dare_init: initialisation for the DARE solver
    :return:
        ell: the log-marginal likelihood log p(y), for hyperparameter optimisation (learning) [scalar]
        means: intermediate filtering means [N, state_dim, 1]
        covs: intermediate filtering covariances [N, state_dim, state_dim]
    """
    if mask is None:
        mask = np.zeros_like(y, dtype=bool)
    Pinf = kernel.stationary_covariance()
    minf = np.zeros([Pinf.shape[0], 1])
    A = kernel.state_transition(dt[1])
    Q = process_noise_covariance(A, Pinf)
    H = kernel.measurement_model()

    # solve the discrete algebraic Ricatti equation to compute the steady-state covariance
    dare_init = Pinf if dare_init is None else dare_init
    Pdare = dare(A, H, Q, noise_cov_tied, dare_init, dare_iters)

    # innovation variance
    S = H @ Pdare @ H.T + noise_cov_tied
    # stationary gain
    K = Pdare @ solve(S, H).T
    # precalculate
    HA = H @ A

    N = dt.shape[0]
    if heteroscedastic:
        # these operations are O(Ns^3) (s=num spatial pts), but can be run in parallel
        Ss = H @ Pdare @ H.T + noise_cov
        Ks = Pdare @ transpose(solve(Ss, np.tile(H, [N, 1, 1])))  # <-- bottleneck
        if not parallel:
            AKHAs = A - Ks @ HA
    else:
        # the homoscedastic version (Gaussian likelihood & no missing data) is O(Nd^2)
        Ss = np.tile(S, [N, 1, 1])
        Ks = np.tile(K, [N, 1, 1])
        AKHAs = np.tile(A - K @ HA, [N, 1, 1])

    # precalculate
    Kys = Ks @ y

    # infinite-horizon state covariance
    cov = Pdare - K @ H @ Pdare

    if parallel:
        if heteroscedastic:
            iA = kernel.state_transition(-dt[1])  # inv(A)
            iAKHAs = iA + iA @ ((Pdare @ H.T) @ solve(noise_cov, np.tile(H, [N, 1, 1])))  # inv(AKHAs)
            ell, means = _parallel_kf_ih(iAKHAs, HA, Kys, Ss, y, minf, mask, heteroscedastic)
        else:
            # AKHAs = A - Ks @ HA
            ell, means = _parallel_kf_ih(AKHAs, HA, Kys, Ss, y, minf, mask, heteroscedastic)
    else:
        ell, means = _sequential_kf_ih(AKHAs, HA, Kys, Ss, y, minf, mask)

    covs = (Pdare, cov)

    return ell, (means, covs)


def rts_dare(A, Q, Pinf, num_iters=20):
    """
    solve a discrete algebraic Ricatti equation
    :param A: transition matrix
    :param Q: process noise covariance
    :param Pinf: initialisation
    :param num_iters: number of iterations
    :return:
    """

    def body(X, _):
        X = A @ X @ A.T + Q
        return X, X

    X_final, _ = scan(f=body,
                      init=Pinf,
                      xs=np.zeros(num_iters))

    return X_final


def _sequential_rts_ih(fms, Afms, H, gain, return_full):

    def body(sm, inputs):
        fm, Afm = inputs

        sm = fm + gain @ (sm - Afm)

        if return_full:
            return sm, sm
        else:
            return sm, H @ sm

    _, sms = scan(f=body,
                  init=fms[-1],
                  xs=(fms, Afms),
                  reverse=True)
    return sms


@vmap
def parallel_smoothing_operator_ih(elem1, elem2):
    B1, v1 = elem1
    B2, v2 = elem2
    return B1 @ B2, B2 @ v1 + v2


def _parallel_rts_ih(fms, Afms, H, gain, return_full):

    gains = np.tile(gain, [fms.shape[0], 1, 1])
    smoothing_elems = (gains, fms - gains @ Afms)
    last_elems = (np.zeros_like(gain), fms[-1])
    initial_elements = tuple(np.append(gen_el, np.expand_dims(last_el, 0), axis=0)
                             for gen_el, last_el in zip(smoothing_elems, last_elems))
    final_elements = associative_scan(parallel_smoothing_operator_ih, initial_elements, reverse=True)
    sms = final_elements[1][:-1]

    if return_full:
        return sms
    else:
        return H @ sms


def rauch_tung_striebel_smoother_infinite_horizon(dt, kernel, filter_mean, filter_cov, return_full=False,
                                                  parallel=False, dare_iters=20, dare_init=None):
    """
    Run the RTS smoother to get p(fₙ|y₁,...,y_N),
    :param dt: step sizes [N, 1]
    :param kernel: an instantiation of the kernel class, used to determine the state space model
    :param filter_mean: the intermediate distribution means computed during filtering [N, state_dim, 1]
    :param filter_cov: the intermediate distribution covariances computed during filtering [N, state_dim, state_dim]
    :param return_full: a flag determining whether to return the full state distribution or just the function(s)
    :param parallel: flag to switch between parallel and sequential implementation of smoother
    :param dare_iters: number of iterations to run the DARE solver for
    :param dare_init: initialisation for the DARE solver
    :return:
        smoothed_mean: the posterior marginal means [N, obs_dim]
        smoothed_var: the posterior marginal variances [N, obs_dim]
    """
    Pinf = kernel.stationary_covariance()
    A = kernel.state_transition(dt[0])
    H = kernel.measurement_model()
    N = dt.shape[0]

    (Pdare, filter_cov) = filter_cov  # unpack filter result
    state_dim = A.shape[1]
    if Pdare.shape[0] != state_dim:  # the sparse Markov case
        Pdare = Pdare[state_dim:, state_dim:]

    # the stationary gain
    gain = filter_cov @ solve(Pdare, A).T

    # solve the Riccati equation
    Qdare = filter_cov - gain @ Pdare @ gain.T
    # Qdare = (Qdare + transpose(Qdare)) / 2
    dare_init = Pinf if dare_init is None else dare_init
    dare_cov = rts_dare(gain, Qdare, dare_init, dare_iters)

    # pre-compute
    Afms = A @ filter_mean

    if parallel:
        means = _parallel_rts_ih(filter_mean, Afms, H, gain, return_full)
    else:
        means = _sequential_rts_ih(filter_mean, Afms, H, gain, return_full)

    if return_full:
        cov = dare_cov
    else:
        cov = H @ dare_cov @ H.T
    covs = np.tile(cov, [N, 1, 1])
    gains = np.tile(gain, [N, 1, 1])

    return means, covs, gains, dare_cov


def dare_pairs(A, Q, R, Pinit, num_iters=20):
    """
    solve a discrete algebraic Ricatti equation
    :param A: transition matrix
    :param Q: process noise covariance
    :param R: observation noise covariance
    :param Pinit: initialisation
    :param num_iters: number of iterations
    :return:
    """
    state_dim = A.shape[0]

    def body(X, _):
        S = X + R
        K = solve(S, X).T
        X = X - K @ X  # update
        X_left = X[state_dim:, state_dim:]
        X_right = A @ X_left @ A.T + Q  # predict
        XA_ = X_left @ A.T
        X = np.block([[X_left, XA_],
                      [XA_.T,  X_right]])
        return X, X

    X_final, _ = scan(f=body,
                      init=Pinit,
                      xs=np.zeros(num_iters))

    return X_final


def _sequential_kf_ih_pairs(IKs, A, Kys, Ss, ys, m0):
    state_dim = A.shape[0]

    def body(carry, inputs):
        y, IK, Ky, S = inputs
        m_left, ell = carry

        # predict
        m_right = A @ m_left

        # construct the joint distribution p(uₙ₋₁,uₙ) = p(uₙ₋₁)p(uₙ|uₙ₋₁)
        m_joint = np.block([[m_left],
                            [m_right]])

        ell_n = mvn_logpdf(y, m_joint, S)
        ell = ell + ell_n

        # perform update
        m = IK @ m_joint + Ky

        # marginalise and store the now fully updated left state, uₙ₋₁
        m_left = m[:state_dim]
        # marginalise and propagate the right state, uₙ
        m_right = m[state_dim:]

        return (m_right, ell), m_left

    (_, loglik), fms = scan(f=body,
                            init=(m0, 0.),
                            xs=(ys, IKs, Kys, Ss))
    return loglik, fms[1:]  # discard intial dummy state


def kalman_filter_infinite_horizon_pairs(dt, kernel, y, noise_cov, parallel=False,
                                         noise_cov_tied=None, dare_iters=20, dare_init=None):
    """
    A Kalman filter over pairs of states, in which y is [2D, 1] and noise_cov is [2D, 2D]
    :param dt: step sizes [N, 1]
    :param kernel: an instantiation of the kernel class, used to determine the state space model
    :param y: observations [N, D, 1]
    :param noise_cov: observation noise covariances [N, D, D]
    :param parallel: flag to switch between parallel and sequential implementation of Kalman filter
    :param noise_cov_tied: averaged observation noise covariance [N, D, D]
    :param dare_iters: number of iterations to run the DARE solver for
    :param dare_init: initialisation for the DARE solver
    :return:
        ell: the log-marginal likelihood log p(y), for hyperparameter optimisation (learning) [scalar]
        means: intermediate filtering means [N, state_dim, 1]
        covs: intermediate filtering covariances [N, state_dim, state_dim]
    """

    Pinf = kernel.stationary_covariance()
    minf = np.zeros([Pinf.shape[0], 1])

    A = kernel.state_transition(dt[1])
    Q = process_noise_covariance(A, Pinf)

    # solve the discrete algebraic Ricatti equation to compute the steady-state covariance
    dare_init = Pinf if dare_init is None else dare_init
    Pdare = dare_pairs(A, Q, noise_cov_tied, dare_init, dare_iters)

    # innovation variance
    S = Pdare + noise_cov_tied
    # stationary gain
    K = solve(S, Pdare).T

    # these operations are O(Nd^3), but can be run in parallel, so the full algorith would be O(Nd^2) on a GPU
    Ss = Pdare + noise_cov
    Ks = transpose(vmap(solve, [0, None])(Ss, Pdare))  # <-- bottleneck
    IKs = np.eye(K.shape[0]) - Ks

    # precalculate
    Kys = Ks @ y

    # infinite-horizon state covariance
    cov = Pdare - K @ Pdare

    if parallel:
        raise NotImplementedError("Parallel KF not implemented yet")
    else:
        ell, means = _sequential_kf_ih_pairs(IKs, A, Kys, Ss, y, minf)

    state_dim = A.shape[0]
    covs = (Pdare, cov[:state_dim, :state_dim])

    return ell, (means, covs)



import jax.numpy as np
import numpy as nnp
from jax import vmap
from jax.ops import index_add, index, index_update
from jax.scipy.linalg import cholesky, cho_factor, cho_solve
from jax.scipy.special import gammaln
from jax.lax import scan
# from matplotlib._png import read_png
import math
from functools import partial
import numba as nb
import numpy as onp

LOG2PI = math.log(2 * math.pi)
INV2PI = (2 * math.pi) ** -1


def solve(P, Q):
    """
    Compute P^-1 Q, where P is a PSD matrix, using the Cholesky factorisation
    """
    L = cho_factor(P, lower=True)
    return cho_solve(L, Q)


def inv(P):
    """
    Compute the inverse of a PSD matrix using the Cholesky factorisation
    """
    L = cho_factor(P, lower=True)
    return cho_solve(L, np.eye(P.shape[-1]))


def inv_vmap(P):
    """
    Compute the inverse of a PSD matrix using the Cholesky factorisation
    """
    L = cho_factor(P, lower=True)
    return cho_solve(L, np.tile(np.eye(P.shape[-1]), [P.shape[0], 1, 1]))


def diag(P):
    """
    a broadcastable version of np.diag, for when P is size [N, D, D]
    """
    return np.diagonal(P, axis1=1, axis2=2)


@vmap
def vmap_diag(P):
    return np.diag(P)


def transpose(P):
    return np.swapaxes(P, -1, -2)


def softplus(x_):
    return np.log(1. + np.exp(x_))
    # return np.log(1. + np.exp(-np.abs(x_))) + np.maximum(x_, 1e-24)  # safer version (but derivatve can have issues)


def sigmoid(x_):
    return np.exp(x_) / (np.exp(x_) + 1.)


def sigmoid_diff(x_):
    return sigmoid(x_) * (1. - sigmoid(x_))


def softplus_inv(x_):
    """
    Inverse of the softplus positiviy mapping, used for transforming parameters.
    """
    if x_ is None:
        return x_
    else:
        return np.log(np.exp(x_) - 1)
        # return np.log(1. - np.exp(-np.abs(x_))) + np.maximum(x_, 1e-24)  # safer version


def ensure_positive_precision(K):
    """
    Check whether matrix K has positive diagonal elements.
    If not, then replace the negative elements with default value 0.01
    """
    # K_diag = diag(diag(K))
    K_diag = vmap_diag(diag(K))
    K = np.where(np.any(diag(K) < 0), np.where(K_diag < 0, 1e-2, K_diag), K)
    return K


def ensure_diagonal_positive_precision(K):
    """
    Return a diagonal matrix with all positive values.
    """
    # K_diag = diag(K)[..., None] * np.eye(K.shape[1])
    K_diag = vmap_diag(diag(K))
    K = np.where(K_diag < 0, 1e-2, K_diag)
    return K


def predict_from_state(x_test, ind, x, post_mean, post_cov, gain, kernel):
    """
    wrapper function to vectorise predict_at_t_()
    """
    predict_from_state_func = vmap(
        predict_from_state_, (0, 0, None, None, None, None, None)
    )
    return predict_from_state_func(x_test, ind, x, post_mean, post_cov, gain, kernel)


def predict_from_state_(x_test, ind, x, post_mean, post_cov, gain, kernel):
    """
    predict the state distribution at time t by projecting from the neighbouring inducing states
    """
    P, T = compute_conditional_statistics(x_test, x, kernel, ind)
    # joint posterior (i.e. smoothed) mean and covariance of the states [u_, u+] at time t:
    mean_joint = np.block([[post_mean[ind]],
                           [post_mean[ind + 1]]])
    cross_cov = gain[ind] @ post_cov[ind + 1]
    cov_joint = np.block([[post_cov[ind], cross_cov],
                          [cross_cov.T, post_cov[ind + 1]]])
    return P @ mean_joint, P @ cov_joint @ P.T + T


def temporal_conditional(X, X_test, mean, cov, gain, kernel):
    """
    predict from time X to time X_test give state mean and covariance at X
    """
    Pinf = kernel.stationary_covariance()[None, ...]
    minf = np.zeros([1, Pinf.shape[1], 1])
    mean_aug = np.concatenate([minf, mean, minf])
    cov_aug = np.concatenate([Pinf, cov, Pinf])
    gain = np.concatenate([np.zeros_like(gain[:1]), gain])

    # figure out which two training states each test point is located between
    ind_test = np.searchsorted(X.reshape(-1, ), X_test.reshape(-1, )) - 1

    # project from training states to test locations
    test_mean, test_cov = predict_from_state(X_test, ind_test, X, mean_aug, cov_aug, gain, kernel)

    return test_mean, test_cov


def predict_from_state_infinite_horizon(x_test, ind, x, post_mean, kernel):
    """
    wrapper function to vectorise predict_at_t_()
    """
    predict_from_state_func = vmap(
        predict_from_state_infinite_horizon_, (0, 0, None, None, None)
    )
    return predict_from_state_func(x_test, ind, x, post_mean, kernel)


def predict_from_state_infinite_horizon_(x_test, ind, x, post_mean, kernel):
    """
    predict the state distribution at time t by projecting from the neighbouring inducing states
    """
    P, T = compute_conditional_statistics(x_test, x, kernel, ind)
    # joint posterior (i.e. smoothed) mean and covariance of the states [u_, u+] at time t:
    mean_joint = np.block([[post_mean[ind]],
                           [post_mean[ind + 1]]])
    return P @ mean_joint


def temporal_conditional_infinite_horizon(X, X_test, mean, cov, gain, kernel):
    """
    predict from time X to time X_test give state mean and covariance at X
    """
    Pinf = kernel.stationary_covariance()[None, ...]
    minf = np.zeros([1, Pinf.shape[1], 1])
    mean_aug = np.concatenate([minf, mean, minf])

    # figure out which two training states each test point is located between
    ind_test = np.searchsorted(X.reshape(-1, ), X_test.reshape(-1, )) - 1

    # project from training states to test locations
    test_mean = predict_from_state_infinite_horizon(X_test, ind_test, X, mean_aug, kernel)

    return test_mean, np.tile(cov[0], [test_mean.shape[0], 1, 1])


def compute_conditional_statistics(x_test, x, kernel, ind):
    """
    This version uses cho_factor and cho_solve - much more efficient when using JAX

    Predicts marginal states at new time points. (new time points should be sorted)
    Calculates the conditional density:
             p(xₙ|u₋, u₊) = 𝓝(Pₙ @ [u₋, u₊], Tₙ)

    :param x_test: time points to generate observations for [N]
    :param x: inducing state input locations [M]
    :param kernel: prior object providing access to state transition functions
    :param ind: an array containing the index of the inducing state to the left of every input [N]
    :return: parameters for the conditional mean and covariance
            P: [N, D, 2*D]
            T: [N, D, D]
    """
    dt_fwd = x_test[..., 0] - x[ind, 0]
    dt_back = x[ind + 1, 0] - x_test[..., 0]
    A_fwd = kernel.state_transition(dt_fwd)
    A_back = kernel.state_transition(dt_back)
    Pinf = kernel.stationary_covariance()
    Q_fwd = Pinf - A_fwd @ Pinf @ A_fwd.T
    Q_back = Pinf - A_back @ Pinf @ A_back.T
    A_back_Q_fwd = A_back @ Q_fwd
    Q_mp = Q_back + A_back @ A_back_Q_fwd.T

    jitter = 1e-8 * np.eye(Q_mp.shape[0])
    chol_Q_mp = cho_factor(Q_mp + jitter, lower=True)
    Q_mp_inv_A_back = cho_solve(chol_Q_mp, A_back)  # V = Q₋₊⁻¹ Aₜ₊

    # The conditional_covariance T = Q₋ₜ - Q₋ₜAₜ₊ᵀQ₋₊⁻¹Aₜ₊Q₋ₜ == Q₋ₜ - Q₋ₜᵀAₜ₊ᵀL⁻ᵀL⁻¹Aₜ₊Q₋ₜ
    T = Q_fwd - A_back_Q_fwd.T @ Q_mp_inv_A_back @ Q_fwd
    # W = Q₋ₜAₜ₊ᵀQ₋₊⁻¹
    W = Q_fwd @ Q_mp_inv_A_back.T
    P = np.concatenate([A_fwd - W @ A_back @ A_fwd, W], axis=-1)
    return P, T


def sum_natural_params_by_group(carry, inputs):
    ind_m, nat1_m, nat2_m = inputs
    nat1s, nat2s, count = carry
    nat1s = index_add(nat1s, index[ind_m], nat1_m)
    nat2s = index_add(nat2s, index[ind_m], nat2_m)
    count = index_add(count, index[ind_m], 1.0)
    return (nat1s, nat2s, count), 0.


def count_indices(carry, inputs):
    ind_m = inputs
    count = carry
    count = index_add(count, index[ind_m], 1.0)
    return count, 0.


def input_admin(t, y, r):
    """
    Order the inputs.
    :param t: training inputs [N, 1]
    :param y: observations at the training inputs [N, 1]
    :param r: training spatial inputs
    :return:
        t_train: training inputs [N, 1]
        y_train: training observations [N, R]
        r_train: training spatial inputs [N, R]
        dt_train: training step sizes, Δtₙ = tₙ - tₙ₋₁ [N, 1]
    """
    assert t.shape[0] == y.shape[0]
    if t.ndim < 2:
        t = nnp.expand_dims(t, 1)  # make 2-D
    if y.ndim < 2:
        y = nnp.expand_dims(y, 1)  # make 2-D
    if r is None:
        if t.shape[1] > 1:
            r = t[:, 1:]
            t = t[:, :1]
        else:
            r = nnp.nan * t  # np.empty((1,) + x.shape[1:]) * np.nan
    if r.ndim < 2:
        r = nnp.expand_dims(r, 1)  # make 2-D
    ind = nnp.argsort(t[:, 0], axis=0)
    t_train = t[ind, ...]
    y_train = y[ind, ...]
    r_train = r[ind, ...]
    dt_train = nnp.concatenate([np.array([0.0]), nnp.diff(t_train[:, 0])])
    return (np.array(t_train, dtype=np.float64), np.array(y_train, dtype=np.float64),
            np.array(r_train, dtype=np.float64), np.array(dt_train, dtype=np.float64))


def create_spatiotemporal_grid(X, Y):
    """
    create a grid of data sized [T, R1, R2]
    note that this function removes full duplicates (i.e. where all dimensions match)
    TODO: generalise to >5D
    """
    if Y.ndim < 2:
        Y = Y[:, None]
    num_spatial_dims = X.shape[1] - 1
    if num_spatial_dims == 4:
        sort_ind = nnp.lexsort((X[:, 4], X[:, 3], X[:, 2], X[:, 1], X[:, 0]))  # sort by 0, 1, 2, 4
    elif num_spatial_dims == 3:
        sort_ind = nnp.lexsort((X[:, 3], X[:, 2], X[:, 1], X[:, 0]))  # sort by 0, 1, 2, 3
    elif num_spatial_dims == 2:
        sort_ind = nnp.lexsort((X[:, 2], X[:, 1], X[:, 0]))  # sort by 0, 1, 2
    elif num_spatial_dims == 1:
        sort_ind = nnp.lexsort((X[:, 1], X[:, 0]))  # sort by 0, 1
    else:
        raise NotImplementedError
    X = X[sort_ind]
    Y = Y[sort_ind]
    unique_time = np.unique(X[:, 0])
    unique_space = nnp.unique(X[:, 1:], axis=0)
    N_t = unique_time.shape[0]
    N_r = unique_space.shape[0]
    if num_spatial_dims == 4:
        R = np.tile(unique_space, [N_t, 1, 1, 1, 1])
    elif num_spatial_dims == 3:
        R = np.tile(unique_space, [N_t, 1, 1, 1])
    elif num_spatial_dims == 2:
        R = np.tile(unique_space, [N_t, 1, 1])
    elif num_spatial_dims == 1:
        R = np.tile(unique_space, [N_t, 1])
    else:
        raise NotImplementedError
    R_flat = R.reshape(-1, num_spatial_dims)
    Y_dummy = np.nan * np.zeros([N_t * N_r, 1])
    time_duplicate = np.tile(unique_time, [N_r, 1]).T.flatten()
    X_dummy = np.block([time_duplicate[:, None], R_flat])
    X_all = np.vstack([X, X_dummy])
    Y_all = np.vstack([Y, Y_dummy])
    X_unique, ind = nnp.unique(X_all, axis=0, return_index=True)
    Y_unique = Y_all[ind]
    grid_shape = (unique_time.shape[0], ) + unique_space.shape
    R_grid = X_unique[:, 1:].reshape(grid_shape)
    Y_grid = Y_unique.reshape(grid_shape[:-1] + (1, ))
    return unique_time[:, None], R_grid, Y_grid


def discretegrid(xy, w, nt):
    """
    Convert spatial observations to a discrete intensity grid
    :param xy: observed spatial locations as a two-column vector
    :param w: observation window, i.e. discrete grid to be mapped to, [xmin xmax ymin ymax]
    :param nt: two-element vector defining number of bins in both directions
    """
    # Make grid
    x = nnp.linspace(w[0], w[1], nt[0] + 1)
    y = nnp.linspace(w[2], w[3], nt[1] + 1)
    X, Y = nnp.meshgrid(x, y)

    # Count points
    N = nnp.zeros([nt[1], nt[0]])
    for i in range(nt[0]):
        for j in range(nt[1]):
            ind = (xy[:, 0] >= x[i]) & (xy[:, 0] < x[i + 1]) & (xy[:, 1] >= y[j]) & (xy[:, 1] < y[j + 1])
            N[j, i] = nnp.sum(ind)
    return X[:-1, :-1].T, Y[:-1, :-1].T, N.T


def gaussian_log_expected_lik(y, post_mean, post_cov, obs_var):
    """
    Calculates the log partition function:
        logZₙ = log ∫ 𝓝(yₙ|fₙ,σ²) 𝓝(fₙ|mₙ,vₙ) dfₙ = E[𝓝(yₙ|fₙ,σ²)]
    :param y: data / observation (yₙ)
    :param post_mean: posterior mean (mₙ)
    :param post_cov: posterior variance (vₙ)
    :param obs_var: variance, σ², of the Gaussian observation model p(yₙ|fₙ)=𝓝(yₙ|fₙ,σ²)
    :return:
        lZ: the log partition function, logZₙ [scalar]
    """
    post_mean = post_mean.reshape(-1, 1)
    post_var = np.diag(post_cov).reshape(-1, 1)
    y = y.reshape(-1, 1)
    obs_var = np.diag(obs_var).reshape(-1, 1)

    var = obs_var + post_var
    prec = 1 / var
    # version which computes sum and outputs scalar
    # lZ = (
    #     -0.5 * y.shape[-2] * np.log(2 * np.pi)
    #     - 0.5 * np.sum((y - post_mean) * prec * (y - post_mean))
    #     - 0.5 * np.sum(np.log(np.maximum(var, 1e-10)))
    # )
    # version which computes individual parts and outputs vector
    lZ = (
        -0.5 * np.log(2 * np.pi)
        - 0.5 * (y - post_mean) * prec * (y - post_mean)
        - 0.5 * np.log(np.maximum(var, 1e-10))
    )
    return lZ


def log_chol_matrix_det(chol):
    val = np.square(np.diag(chol))
    return np.sum(np.log(val))


def mvn_logpdf(x, mean, cov, mask=None):
    """
    evaluate a multivariate Gaussian (log) pdf
    """
    x = x.reshape(-1, 1)
    mean = mean.reshape(-1, 1)
    if mask is not None:
        # build a mask for computing the log likelihood of a partially observed multivariate Gaussian
        maskv = mask.reshape(-1, 1)
        x = np.where(maskv, 0., x)
        mean = np.where(maskv, 0., mean)
        cov_masked = np.where(maskv + maskv.T, 0., cov)  # ensure masked entries are independent
        cov = np.where(np.diag(mask), INV2PI, cov_masked)  # ensure masked entries return log like of 0

    n = mean.shape[0]
    cho, low = cho_factor(cov, lower=True)
    log_det = 2 * np.sum(np.log(np.abs(np.diag(cho))))
    diff = x - mean
    scaled_diff = cho_solve((cho, low), diff)
    distance = diff.T @ scaled_diff
    return np.squeeze(-0.5 * (distance + n * LOG2PI + log_det))


def mvst_logpdf(x, mean, scale, df, mask=None):
    """
    evaluate a multivariate Student's t (log) pdf
    """
    x = x.reshape(-1, 1)
    mean = mean.reshape(-1, 1)
    dim = scale.shape[0]
    if mask is not None:
        maskv = mask.reshape(-1, 1)
        # build a mask for computing the log likelihood of a partially observed multivariate Student's t
        x = np.where(maskv, 0., x)
        mean = np.where(maskv, 0., mean)
        scale_masked = np.where(maskv + maskv.T, 0., scale)  # ensure masked entries are independent
        scale = np.where(np.diag(mask), 1, scale_masked)  # ensure masked entries return log like of 0

    log_det = 2 * np.sum(np.log(np.abs(np.diag(scale))))
    const = (
            gammaln((df + dim) * 0.5)
            - gammaln(df * 0.5)
            - 0.5 * dim * (np.log(df) + np.log(np.pi))
            - 0.5 * log_det
    )
    return const - 0.5 * (df + dim) * np.log(
        1.0 + (1.0 / df) * np.squeeze((x - mean).T @ solve(scale, (x - mean)))
    )


def pep_constant(var, power, mask=None):
    dim = var.shape[1]
    chol = cholesky(var, lower=True)
    log_diag_chol = np.log(np.abs(np.diag(chol)))

    if mask is not None:
        log_diag_chol = np.where(mask, 0., log_diag_chol)
        dim -= np.sum(np.array(mask, dtype=int))

    logdetvar = 2 * np.sum(log_diag_chol)
    constant = (
        0.5 * dim * ((1 - power) * LOG2PI - np.log(power))
        + 0.5 * (1 - power) * logdetvar
    )
    return constant


def mvn_logpdf_and_derivs(x, mean, cov, mask=None):
    """
    evaluate a multivariate Gaussian (log) pdf and compute its derivatives w.r.t. the mean
    """
    if mask is not None:
        # build a mask for computing the log likelihood of a partially observed multivariate Gaussian
        maskv = mask.reshape(-1, 1)
        mean = np.where(maskv, x, mean)
        cov_masked = np.where(maskv + maskv.T, 0., cov)  # ensure masked entries are independent
        cov = np.where(np.diag(mask), INV2PI, cov_masked)  # ensure masked entries return log like of 0

    n = mean.shape[0]
    cho, low = cho_factor(cov, lower=True)
    precision = cho_solve((cho, low), np.eye(cho.shape[1]))  # second derivative
    log_det = 2 * np.sum(np.log(np.abs(np.diag(cho))))
    diff = x - mean
    scaled_diff = precision @ diff  # first derivative
    distance = diff.T @ scaled_diff
    return np.squeeze(-0.5 * (distance + n * LOG2PI + log_det)), scaled_diff, -precision


@vmap
def _gaussian_expected_log_lik(y, post_mean, post_cov, var):
    post_mean = post_mean.reshape(-1, 1)
    post_cov = post_cov.reshape(-1, 1)
    y = y.reshape(-1, 1)
    var = var.reshape(-1, 1)
    # version which computes sum and outputs scalar
    # exp_log_lik = (
    #     -0.5 * y.shape[-2] * np.log(2 * np.pi)  # multiplier based on dimensions needed if taking sum of other terms
    #     - 0.5 * np.sum(np.log(var))
    #     - 0.5 * np.sum(((y - post_mean) ** 2 + post_cov) / var)
    # )
    # version which computes individual parts and outputs vector
    exp_log_lik = (
        -0.5 * np.log(2 * np.pi)
        - 0.5 * np.log(var)
        - 0.5 * ((y - post_mean) ** 2 + post_cov) / var
    )
    return exp_log_lik


def gaussian_expected_log_lik_diag(y, post_mean, post_cov, var):
    """
    Computes the "variational expectation", i.e. the
    expected log-likelihood, and its derivatives w.r.t. the posterior mean
        E[log 𝓝(yₙ|fₙ,σ²)] = ∫ log 𝓝(yₙ|fₙ,σ²) 𝓝(fₙ|mₙ,vₙ) dfₙ
    :param y: data / observation (yₙ)
    :param post_mean: posterior mean (mₙ)
    :param post_cov: posterior variance (vₙ)
    :param var: variance, σ², of the Gaussian observation model p(yₙ|fₙ)=𝓝(yₙ|fₙ,σ²)
    :return:
        exp_log_lik: the expected log likelihood, E[log 𝓝(yₙ|fₙ,var)]  [scalar]
    """
    post_cov = diag(post_cov)
    var = diag(var)
    # var_exp = vmap(_gaussian_expected_log_lik)(y, post_mean, post_cov, var)
    var_exp = _gaussian_expected_log_lik(y, post_mean, post_cov, var)
    # return np.sum(var_exp)
    return var_exp


def gaussian_expected_log_lik(Y, q_mu, q_covar, noise, mask=None):
    """
    :param Y: N x 1
    :param q_mu: N x 1
    :param q_covar: N x N
    :param noise: N x N
    :param mask: N x 1
    :return:
        E[log 𝓝(yₙ|fₙ,σ²)] = ∫ log 𝓝(yₙ|fₙ,σ²) 𝓝(fₙ|mₙ,vₙ) dfₙ
    """

    if mask is not None:
        # build a mask for computing the log likelihood of a partially observed multivariate Gaussian
        maskv = mask.reshape(-1, 1)
        q_mu = np.where(maskv, Y, q_mu)
        noise = np.where(maskv + maskv.T, 0., noise)  # ensure masked entries are independent
        noise = np.where(np.diag(mask), INV2PI, noise)  # ensure masked entries return log like of 0
        q_covar = np.where(maskv + maskv.T, 0., q_covar)  # ensure masked entries are independent
        q_covar = np.where(np.diag(mask), 1e-20, q_covar)  # ensure masked entries return trace term of 0

    ml = mvn_logpdf(Y, q_mu, noise)
    trace_term = -0.5 * np.trace(solve(noise, q_covar))
    return ml + trace_term


def compute_cavity(post_mean, post_cov, site_nat1, site_nat2, power, jitter=1e-8):
    """
    remove local likelihood approximation  from the posterior to obtain the marginal cavity distribution
    """
    post_nat2 = inv(post_cov + jitter * np.eye(post_cov.shape[1]))
    cav_cov = inv(post_nat2 - power * site_nat2)  # cavity covariance
    cav_mean = cav_cov @ (post_nat2 @ post_mean - power * site_nat1)  # cavity mean
    return cav_mean, cav_cov


def build_joint(ind, mean, cov, smoother_gain):
    """
    joint posterior (i.e. smoothed) mean and covariance of the states [u_, u+] at time t
    """
    mean_joint = np.block([[mean[ind]],
                           [mean[ind + 1]]])
    cross_cov = smoother_gain[ind] @ cov[ind + 1]
    cov_joint = np.block([[cov[ind], cross_cov],
                          [cross_cov.T, cov[ind + 1]]])
    return mean_joint, cov_joint


def set_z_stats(t, z):
    ind = (np.searchsorted(z.reshape(-1, ), t[:, :1].reshape(-1, )) - 1)
    num_neighbours = np.array([np.sum(ind == m) for m in range(z.shape[0] - 1)])
    return ind, num_neighbours


def gaussian_first_derivative_wrt_mean(f, m, C, w):
    invC = inv(C)
    return invC @ (f - m) * w


def gaussian_second_derivative_wrt_mean(f, m, C, w):
    invC = inv(C)
    return (invC @ (f - m) @ (f - m).T @ invC - invC) * w


def scaled_squared_euclid_dist(X, X2, ell):
    """
    Returns ‖(X - X2ᵀ) / ℓ‖², i.e. the squared L₂-norm.
    Adapted from GPflow: https://github.com/GPflow/GPflow
    """
    return square_distance(X / ell, X2 / ell)


def square_distance(X, X2):
    """
    Adapted from GPflow: https://github.com/GPflow/GPflow

    Returns ||X - X2ᵀ||²
    Due to the implementation and floating-point imprecision, the
    result may actually be very slightly negative for entries very
    close to each other.

    This function can deal with leading dimensions in X and X2.
    In the sample case, where X and X2 are both 2 dimensional,
    for example, X is [N, D] and X2 is [M, D], then a tensor of shape
    [N, M] is returned. If X is [N1, S1, D] and X2 is [N2, S2, D]
    then the output will be [N1, S1, N2, S2].
    """
    Xs = np.sum(np.square(X), axis=-1)
    X2s = np.sum(np.square(X2), axis=-1)
    dist = -2 * np.tensordot(X, X2, [[-1], [-1]])
    dist += broadcasting_elementwise(np.add, Xs, X2s)
    return dist


def broadcasting_elementwise(op, a, b):
    """
    Adapted from GPflow: https://github.com/GPflow/GPflow

    Apply binary operation `op` to every pair in tensors `a` and `b`.

    :param op: binary operator on tensors, e.g. tf.add, tf.substract
    :param a: tf.Tensor, shape [n_1, ..., n_a]
    :param b: tf.Tensor, shape [m_1, ..., m_b]
    :return: tf.Tensor, shape [n_1, ..., n_a, m_1, ..., m_b]
    """
    flatres = op(np.reshape(a, [-1, 1]), np.reshape(b, [1, -1]))
    return flatres.reshape(a.shape[0], b.shape[0])


def rotation_matrix(dt, omega):
    """
    Discrete time rotation matrix
    :param dt: step size [1]
    :param omega: frequency [1]
    :return:
        R: rotation matrix [2, 2]
    """
    R = np.array([
        [np.cos(omega * dt), -np.sin(omega * dt)],
        [np.sin(omega * dt),  np.cos(omega * dt)]
    ])
    return R


@partial(nb.jit, nopython=True)
def nb_balance_ss(F: onp.ndarray,
                  iters: int) -> onp.ndarray:
    """
    taken from https://github.com/EEA-sensors/parallel-gps/blob/main/pssgp/kernels/math_utils.py
    """
    dim = F.shape[0]
    dtype = F.dtype
    d = onp.ones((dim,), dtype=dtype)
    for k in range(iters):
        for i in range(dim):
            tmp = onp.copy(F[:, i])
            tmp[i] = 0.
            c = onp.linalg.norm(tmp, 2)
            tmp2 = onp.copy(F[i, :])
            tmp2[i] = 0.

            r = onp.linalg.norm(tmp2, 2)
            f = onp.sqrt(r / c)
            d[i] *= f
            F[:, i] *= f
            F[i, :] /= f
    return d


def balance(F: np.ndarray,
            iters: int) -> np.ndarray:
    dim = F.shape[0]
    d = np.ones((dim,))

    def loop_over_iters(carry_, _):
        F, d = carry_

        def loop_over_dims(carry, _):
            F, d, i = carry

            tmp = F[:, i]
            tmp = index_update(tmp, index[i], 0.)
            c = np.linalg.norm(tmp, 2)
            tmp2 = F[i, :]
            tmp2 = index_update(tmp2, index[i], 0.)

            r = np.linalg.norm(tmp2, 2)
            f = np.sqrt(r / c)
            d = index_update(d, index[i], d[i] * f)
            F = index_update(F, index[:, i], F[:, i] * f)
            F = index_update(F, index[i, :], F[i, :] / f)

            return (F, d, i+1), d

        (F, d, _), d_all = scan(f=loop_over_dims,
                                init=(F, d, 0),
                                xs=np.zeros(dim))

        return (F, d), d

    (_, d), _ = scan(f=loop_over_iters,
                     init=(F, d),
                     xs=np.zeros(iters))

    return d


# def bitmappify(ax, dpi=None):
#     """
#     # Convert vector axes content to raster (bitmap) images
#     :param ax: the figure axis
#     :param dpi: dots per inch
#     """
#     fig = ax.figure
#     # safe plot without axes
#     ax.set_axis_off()
#     fig.savefig('temp.png', dpi=dpi, transparent=True)
#     ax.set_axis_on()
#     # remember geometry
#     xl = ax.get_xlim()
#     yl = ax.get_ylim()
#     xb = ax.bbox._bbox.corners()[:, 0]
#     xb = (min(xb), max(xb))
#     yb = ax.bbox._bbox.corners()[:, 1]
#     yb = (min(yb), max(yb))
#     # compute coordinates to place bitmap image later
#     xb = (- xb[0] / (xb[1] - xb[0]), (1 - xb[0]) / (xb[1] - xb[0]))
#     xb = (xb[0] * (xl[1] - xl[0]) + xl[0], xb[1] * (xl[1] - xl[0]) + xl[0])
#     yb = (- yb[0] / (yb[1] - yb[0]), (1 - yb[0]) / (yb[1] - yb[0]))
#     yb = (yb[0] * (yl[1] - yl[0]) + yl[0], yb[1] * (yl[1] - yl[0]) + yl[0])
#     # replace the dots by the bitmap
#     del ax.collections[:]
#     del ax.lines[:]
#     ax.imshow(read_png('temp.png'), origin='upper',
#               aspect='auto', extent=(xb[0], xb[1], yb[0], yb[1]), label='_nolegend_')
#     # reset view
#     ax.set_xlim(xl)
#     ax.set_ylim(yl)
